"Serialize & parse WS-Discovery Bye SOAP messages"

import uuid
from ..namespaces import NS_ADDRESSING, NS_DISCOVERY, NS_ACTION_BYE, NS_ADDRESS_ALL
from ..envelope import SoapEnvelope
from ..util import createSkelSoapMessage, getBodyEl, getHeaderEl, addElementWithText, \
                   addTypes, addScopes, getDocAsString, getScopes, addEPR, \
                   _parseAppSequence


def constructBye(service):
    "construct an envelope that represents a ``Bye`` message"

    env = SoapEnvelope()
    env.setAction(NS_ACTION_BYE)
    env.setTo(NS_ADDRESS_ALL)
    env.setMessageId(uuid.uuid4().urn)
    env.setInstanceId(str(service.getInstanceId()))
    env.setMessageNumber(str(service.getMessageNumber()))
    env.setEPR(service.getEPR())
    return env


def createByeMessage(env):
    "serialize a SOAP envelope object into a string"
    doc = createSkelSoapMessage(NS_ACTION_BYE)

    bodyEl = getBodyEl(doc)
    headerEl = getHeaderEl(doc)

    addElementWithText(doc, headerEl, "a:MessageID", NS_ADDRESSING, env.getMessageId())
    addElementWithText(doc, headerEl, "a:To", NS_ADDRESSING, env.getTo())

    appSeqEl = doc.createElementNS(NS_DISCOVERY, "d:AppSequence")
    appSeqEl.setAttribute("InstanceId", env.getInstanceId())
    appSeqEl.setAttribute("MessageNumber", env.getMessageNumber())
    headerEl.appendChild(appSeqEl)

    byeEl = doc.createElementNS(NS_DISCOVERY, "d:Bye")
    addEPR(doc, byeEl, env.getEPR())
    bodyEl.appendChild(byeEl)

    return getDocAsString(doc)


def parseByeMessage(dom):
    "parse a XML message into a SOAP envelope object"
    env = SoapEnvelope()
    env.setAction(NS_ACTION_BYE)

    env.setMessageId(dom.getElementsByTagNameNS(NS_ADDRESSING, "MessageID")[0].firstChild.data.strip())
    env.setTo(dom.getElementsByTagNameNS(NS_ADDRESSING, "To")[0].firstChild.data.strip())

    _parseAppSequence(dom, env)

    env.setEPR(dom.getElementsByTagNameNS(NS_ADDRESSING, "Address")[0].firstChild.data.strip())

    return env






import os
import pytest
from wsdiscovery import WSDiscovery


@pytest.fixture
def wsd():
    "provide the discovery client"
    client = WSDiscovery()
    client.start()
    yield client
    client.stop()


@pytest.fixture
def probe_response():
    "provide a captured SOAP probe response"

    DISCOVER_IP = "192.168.1.104"

    curdir = os.path.dirname(__file__)
    with open(curdir + "/data/probe_response.xml", "rb") as resp:
        canned = resp.read()
    return (canned, DISCOVER_IP)



"""Service scopes are used to constrain service discovery."""


class Scope:
    "Service scope implementation."

    def __init__(self, value, matchBy=None):
        self._matchBy = matchBy
        self._value = value

    def getMatchBy(self):
        return self._matchBy

    def getValue(self):
        return self._value

    def getQuotedValue(self):
        return self._value.replace(' ', '%20')

    def __repr__(self):
        if self.getMatchBy() == None or len(self.getMatchBy()) == 0:
            return self.getValue()
        else:
            return self.getMatchBy() + ":" + self.getValue()





"""Discoverable WS-Discovery service."""
import socket

from .util import _getNetworkAddrs


class Service:
    "A web service representation implementation"

    def __init__(self, types, scopes, xAddrs, epr, instanceId):
        self._types = types
        self._scopes = scopes
        self._xAddrs = xAddrs
        self._epr = epr
        self._instanceId = instanceId
        self._messageNumber = 0
        self._metadataVersion = 1

    def getTypes(self):
        "get service types"
        return self._types

    def setTypes(self, types):
        "set service types"
        self._types = types

    def getScopes(self):
        "get the service scopes"
        return self._scopes

    def setScopes(self, scopes):
        "set the service scopes"
        self._scopes = scopes

    def getXAddrs(self):
        "get service network address"
        ret = []
        ipAddrs = None
        for xAddr in self._xAddrs:
            if '{ip}' in xAddr:
                if ipAddrs is None:
                    ipAddrs = _getNetworkAddrs(socket.AF_INET)
                    ipAddrs.append(_getNetworkAddrs(socket.AF_INET6))
                for ipAddr in ipAddrs:
                    if not ipAddr.is_loopback:
                        ret.append(xAddr.format(ip=str(ipAddr)))
            else:
                ret.append(xAddr)
        return ret

    def setXAddrs(self, xAddrs):
        "set service network address"
        self._xAddrs = xAddrs

    def getEPR(self):
        "get endpoint reference"
        return self._epr

    def setEPR(self, epr):
        "set endpoint reference"
        self._epr = epr

    def getInstanceId(self):
        return self._instanceId

    def setInstanceId(self, instanceId):
        self._instanceId = instanceId

    def getMessageNumber(self):
        return self._messageNumber

    def setMessageNumber(self, messageNumber):
        self._messageNumber = messageNumber

    def getMetadataVersion(self):
        return self._metadataVersion

    def setMetadataVersion(self, metadataVersion):
        self._metadataVersion = metadataVersion

    def incrementMessageNumber(self):
        self._messageNumber = self._messageNumber + 1





"""Discovery application."""

import time
import uuid

from .actions import *
from .uri import URI
from .util import matchesFilter, filterServices, extractSoapUdpAddressFromURI
from .service import Service
from .namespaces import NS_DISCOVERY
from .threaded import ThreadedNetworking
from .daemon import Daemon

DEFAULT_DISCOVERY_TIMEOUT = 3


class Discovery:
    """networking-agnostic generic remote service discovery mixin"""

    def __init__(self, **kwargs):
        self._remoteServices = {}
        self._remoteServiceHelloCallback = None
        self._remoteServiceHelloCallbackTypesFilter = None
        self._remoteServiceHelloCallbackScopesFilter = None
        self._remoteServiceByeCallback = None
        super().__init__(**kwargs)

    def setRemoteServiceHelloCallback(self, cb, types=None, scopes=None):
        """Set callback, which will be called when new service appeared online
        and sent Hi message

        typesFilter and scopesFilter might be list of types and scopes.
        If filter is set, callback is called only for Hello messages,
        which match filter

        Set None to disable callback
        """
        self._remoteServiceHelloCallback = cb
        self._remoteServiceHelloCallbackTypesFilter = types
        self._remoteServiceHelloCallbackScopesFilter = scopes

    def setRemoteServiceByeCallback(self, cb):
        """Set callback, which will be called when new service appeared online
        and sent Hi message
        Service is passed as a parameter to the callback
        Set None to disable callback
        """
        self._remoteServiceByeCallback = cb

    def setRemoteServiceDisappearedCallback(self, cb):
        """Set callback, which will be called when new service disappears
        Service uuid is passed as a parameter to the callback
        Set None to disable callback
        """
        self._remoteServiceDisppearedCallback = cb

    # discovery-related message handlers:

    def _handle_probematches(self, env, addr):
        for match in env.getProbeResolveMatches():
            self._addRemoteService(Service(match.getTypes(), match.getScopes(), match.getXAddrs(), match.getEPR(), 0))
            if match.getXAddrs() is None or len(match.getXAddrs()) == 0:
                self._sendResolve(match.getEPR())

    def _handle_resolvematches(self, env, addr):
        for match in env.getProbeResolveMatches():
            self._addRemoteService(Service(match.getTypes(), match.getScopes(), match.getXAddrs(), match.getEPR(), 0))

    def _handle_hello(self, env, addr):
        #check if it is from a discovery proxy
        rt = env.getRelationshipType()
        if rt is not None and rt.getLocalname() == "Suppression" and rt.getNamespace() == NS_DISCOVERY:
            xAddr = env.getXAddrs()[0]
            #only support 'soap.udp'
            if xAddr.startswith("soap.udp:"):
                self._dpActive = True
                self._dpAddr = extractSoapUdpAddressFromURI(URI(xAddr))
                self._dpEPR = env.getEPR()

        service = Service(env.getTypes(), env.getScopes(), env.getXAddrs(), env.getEPR(), 0)
        self._addRemoteService(service)
        if self._remoteServiceHelloCallback is not None:
            if matchesFilter(service,
                             self._remoteServiceHelloCallbackTypesFilter,
                             self._remoteServiceHelloCallbackScopesFilter):
                self._remoteServiceHelloCallback(service)

    def _handle_bye(self, env, addr):
        #if the bye is from discovery proxy... revert back to multicasting
        if self._dpActive and self._dpEPR == env.getEPR():
            self._dpActive = False
            self._dpAddr = None
            self._dpEPR = None

        self._removeRemoteService(env.getEPR())
        if self._remoteServiceByeCallback is not None:
            self._remoteServiceByeCallback(env.getEPR())


    # handle local address changes:

    def  _networkAddressAdded(self, addr):
        self.addSourceAddr(addr)

    def _networkAddressRemoved(self, addr):
        self.removeSourceAddr(addr)

    # search for & keep track of discovered remote services:

    def _addRemoteService(self, service):
        self._remoteServices[service.getEPR()] = service

    def _removeRemoteService(self, epr):
        if epr in self._remoteServices:
            del self._remoteServices[epr]

    def clearRemoteServices(self):
        'clears remotely discovered services'

        self._remoteServices.clear()

    def searchServices(self, types=None, scopes=None, address=None, port=None,
                       timeout=DEFAULT_DISCOVERY_TIMEOUT):
        'search for services given the TYPES and SCOPES within a given TIMEOUT'
        try:
            self._sendProbe(types, scopes, address, port)
        except:
            raise Exception("Server not started")

        time.sleep(timeout)

        return filterServices(list(self._remoteServices.values()), types, scopes)

    def stop(self):
        self.clearRemoteServices()
        super().stop()


class ThreadedWSDiscovery(Daemon, Discovery, ThreadedNetworking):
    "Full threaded service discovery implementation"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def stop(self):
        super().stop()




"Serialize & parse WS-Discovery Hello SOAP messages"

import uuid
from ..namespaces import NS_ADDRESSING, NS_DISCOVERY, NS_ACTION_HELLO, NS_ADDRESS_ALL
from ..envelope import SoapEnvelope
from ..util import createSkelSoapMessage, getBodyEl, getHeaderEl, addElementWithText, \
                   addTypes, addScopes, getDocAsString, getScopes, getQNameFromValue, \
                   addEPR, addXAddrs, _parseAppSequence, getTypes, getXAddrs


def constructHello(service):
    "construct an envelope that represents a ``Hello`` message"

    service.incrementMessageNumber()

    env = SoapEnvelope()
    env.setAction(NS_ACTION_HELLO)
    env.setTo(NS_ADDRESS_ALL)
    env.setMessageId(uuid.uuid4().urn)
    env.setInstanceId(str(service.getInstanceId()))
    env.setMessageNumber(str(service.getMessageNumber()))
    env.setTypes(service.getTypes())
    env.setScopes(service.getScopes())
    env.setXAddrs(service.getXAddrs())
    env.setEPR(service.getEPR())
    return env


def createHelloMessage(env):
    "serialize a SOAP envelope object into a string"
    doc = createSkelSoapMessage(NS_ACTION_HELLO)

    bodyEl = getBodyEl(doc)
    headerEl = getHeaderEl(doc)

    addElementWithText(doc, headerEl, "a:MessageID", NS_ADDRESSING, env.getMessageId())

    if len(env.getRelatesTo()) > 0:
        addElementWithText(doc, headerEl, "a:RelatesTo", NS_ADDRESSING, env.getRelatesTo())
        relatesToEl = headerEl.getElementsByTagNameNS(NS_ADDRESSING, "RelatesTo")[0]
        relatesToEl.setAttribute("RelationshipType", "d:Suppression")

    addElementWithText(doc, headerEl, "a:To", NS_ADDRESSING, env.getTo())

    appSeqEl = doc.createElementNS(NS_DISCOVERY, "d:AppSequence")
    appSeqEl.setAttribute("InstanceId", env.getInstanceId())
    appSeqEl.setAttribute("MessageNumber", env.getMessageNumber())
    headerEl.appendChild(appSeqEl)

    helloEl = doc.createElementNS(NS_DISCOVERY, "d:Hello")
    addEPR(doc, helloEl, env.getEPR())
    addTypes(doc, helloEl, env.getTypes())
    addScopes(doc, helloEl, env.getScopes())
    addXAddrs(doc, helloEl, env.getXAddrs())
    addElementWithText(doc, helloEl, "d:MetadataVersion", NS_DISCOVERY, env.getMetadataVersion())

    bodyEl.appendChild(helloEl)

    return getDocAsString(doc)


def parseHelloMessage(dom):
    "parse a XML message into a SOAP envelope object"
    env = SoapEnvelope()
    env.setAction(NS_ACTION_HELLO)

    env.setMessageId(dom.getElementsByTagNameNS(NS_ADDRESSING, "MessageID")[0].firstChild.data.strip())
    env.setTo(dom.getElementsByTagNameNS(NS_ADDRESSING, "To")[0].firstChild.data.strip())

    _parseAppSequence(dom, env)

    relatesToNodes = dom.getElementsByTagNameNS(NS_ADDRESSING, "RelatesTo")
    if len(relatesToNodes) > 0:
        env.setRelatesTo(relatesToNodes[0].firstChild.data.strip())
        env.setRelationshipType(getQNameFromValue( \
            relatesToNodes[0].getAttribute("RelationshipType"), relatesToNodes[0]))

    env.setEPR(dom.getElementsByTagNameNS(NS_ADDRESSING, "Address")[0].firstChild.data.strip())

    typeNodes = dom.getElementsByTagNameNS(NS_DISCOVERY, "Types")
    if len(typeNodes) > 0:
        env.setTypes(getTypes(typeNodes[0]))

    scopeNodes = dom.getElementsByTagNameNS(NS_DISCOVERY, "Scopes")
    if len(scopeNodes) > 0:
        env.setScopes(getScopes(scopeNodes[0]))

    xNodes = dom.getElementsByTagNameNS(NS_DISCOVERY, "XAddrs")
    if len(xNodes) > 0:
        env.setXAddrs(getXAddrs(xNodes[0]))

    env.setMetadataVersion(dom.getElementsByTagNameNS(NS_DISCOVERY, "MetadataVersion")[0].firstChild.data.strip())

    return env



"Serialize & parse WS-Discovery Resolve Match SOAP messages"

import uuid

from ..namespaces import NS_ADDRESSING, NS_DISCOVERY, NS_ACTION_RESOLVE_MATCH, NS_ADDRESS_UNKNOWN
from ..envelope import SoapEnvelope
from ..util import createSkelSoapMessage, getBodyEl, getHeaderEl, addElementWithText, \
                   addTypes, getTypes, addScopes, getDocAsString, getScopes, addEPR, \
                   addXAddrs, getXAddrs, _parseAppSequence
                   
from .probematch import ProbeResolveMatch


def constructResolveMatch(service, relatesTo):
    "construct an envelope that represents a ``Resolve Match`` message"

    service.incrementMessageNumber()

    env = SoapEnvelope()
    env.setAction(NS_ACTION_RESOLVE_MATCH)
    env.setTo(NS_ADDRESS_UNKNOWN)
    env.setMessageId(uuid.uuid4().urn)
    env.setInstanceId(str(service.getInstanceId()))
    env.setMessageNumber(str(service.getMessageNumber()))
    env.setRelatesTo(relatesTo)

    prb = ProbeResolveMatch(service.getEPR(), service.getTypes(), service.getScopes(), \
                            service.getXAddrs(), str(service.getMetadataVersion()))
    env.getProbeResolveMatches().append(prb)
    return env


def createResolveMatchMessage(env):
    "serialize a SOAP envelope object into a string"

    doc = createSkelSoapMessage(NS_ACTION_RESOLVE_MATCH)

    bodyEl = getBodyEl(doc)
    headerEl = getHeaderEl(doc)

    addElementWithText(doc, headerEl, "a:MessageID", NS_ADDRESSING, env.getMessageId())
    addElementWithText(doc, headerEl, "a:RelatesTo", NS_ADDRESSING, env.getRelatesTo())
    addElementWithText(doc, headerEl, "a:To", NS_ADDRESSING, env.getTo())

    appSeqEl = doc.createElementNS(NS_DISCOVERY, "d:AppSequence")
    appSeqEl.setAttribute("InstanceId", env.getInstanceId())
    appSeqEl.setAttribute("MessageNumber", env.getMessageNumber())
    headerEl.appendChild(appSeqEl)

    resolveMatchesEl = doc.createElementNS(NS_DISCOVERY, "d:ResolveMatches")
    if len(env.getProbeResolveMatches()) > 0:
        resolveMatch = env.getProbeResolveMatches()[0]
        resolveMatchEl = doc.createElementNS(NS_DISCOVERY, "d:ResolveMatch")
        addEPR(doc, resolveMatchEl, resolveMatch.getEPR())
        addTypes(doc, resolveMatchEl, resolveMatch.getTypes())
        addScopes(doc, resolveMatchEl, resolveMatch.getScopes())
        addXAddrs(doc, resolveMatchEl, resolveMatch.getXAddrs())
        addElementWithText(doc, resolveMatchEl, "d:MetadataVersion", NS_DISCOVERY, resolveMatch.getMetadataVersion())

        resolveMatchesEl.appendChild(resolveMatchEl)

    bodyEl.appendChild(resolveMatchesEl)

    return getDocAsString(doc)


def parseResolveMatchMessage(dom):
    "parse a XML message into a SOAP envelope object"

    env = SoapEnvelope()
    env.setAction(NS_ACTION_RESOLVE_MATCH)

    env.setMessageId(dom.getElementsByTagNameNS(NS_ADDRESSING, "MessageID")[0].firstChild.data.strip())
    env.setRelatesTo(dom.getElementsByTagNameNS(NS_ADDRESSING, "RelatesTo")[0].firstChild.data.strip())
    env.setTo(dom.getElementsByTagNameNS(NS_ADDRESSING, "To")[0].firstChild.data.strip())

    _parseAppSequence(dom, env)

    nodes = dom.getElementsByTagNameNS(NS_DISCOVERY, "ResolveMatch")
    if len(nodes) > 0:
        node = nodes[0]
        epr = node.getElementsByTagNameNS(NS_ADDRESSING, "Address")[0].firstChild.data.strip()

        typeNodes = node.getElementsByTagNameNS(NS_DISCOVERY, "Types")
        types = []
        if len(typeNodes) > 0:
            types = getTypes(typeNodes[0])

        scopeNodes = node.getElementsByTagNameNS(NS_DISCOVERY, "Scopes")
        scopes = []
        if len(scopeNodes) > 0:
            scopes = getScopes(scopeNodes[0])

        xAddrs = getXAddrs(node.getElementsByTagNameNS(NS_DISCOVERY, "XAddrs")[0])
        mdv = node.getElementsByTagNameNS(NS_DISCOVERY, "MetadataVersion")[0].firstChild.data.strip()
        env.getProbeResolveMatches().append(ProbeResolveMatch(epr, types, scopes, xAddrs, mdv))

    return env








"""Qualified name support; see e.g. https://en.wikipedia.org/wiki/QName"""


class QName:
    "Qualified name implementation"

    def __init__(self, namespace, localname, namespace_prefix=None):
        self._namespace = namespace
        self._localname = localname
        self._namespace_prefix = namespace_prefix

    def getNamespace(self):
        return self._namespace

    def getLocalname(self):
        return self._localname

    def getNamespacePrefix(self):
        return self._namespace_prefix

    def getFullname(self):
        return self.getNamespace() + ":" + self.getLocalname()

    def __repr__(self):
        return self.getFullname()






"""Generic networking-agnostic WS-Discovery messaging daemon mixin implementation."""

import random
import time
import uuid
import logging

from .actions import *
from .uri import URI
from .service import Service
from .envelope import SoapEnvelope
from .udp import UNICAST_UDP_REPEAT, MULTICAST_UDP_REPEAT

APP_MAX_DELAY = 500 # miliseconds

logger = logging.getLogger("daemon")


class Daemon:
    "generic WS-Discovery messaging daemon implementation"

    def __init__(self, uuid_=None, capture=None, ttl=1, **kwargs):

        # track existence of a possible discovery proxy
        self._dpActive = False
        self._dpAddr = None
        self._dpEPR = None

        if uuid_ is not None:
            self.uuid = uuid_
        else:
            self.uuid = uuid.uuid4().urn

        self._capture = capture
        self.ttl = ttl
        self._unicast_num = kwargs.get('unicast_num', UNICAST_UDP_REPEAT)
        self._multicast_num = kwargs.get('multicast_num', MULTICAST_UDP_REPEAT)

        super().__init__(**kwargs)

    def envReceived(self, env, addr):
        action = env.getAction()
        action_name = '_handle_' + action[action.rfind('/')+1:].lower()
        try:
            handler = getattr(self, action_name)
        except AttributeError:
            logger.warning("could not find handler for: %s" % action_name)
        else:
            handler(env, addr)

    def _sendResolveMatch(self, service, relatesTo, addr):
        env = constructResolveMatch(service, relatesTo)
        self.sendUnicastMessage(env, addr[0], addr[1], unicast_num=self._unicast_num)

    def _sendProbeMatch(self, services, relatesTo, addr):
        env = constructProbeMatch(services, relatesTo)
        self.sendUnicastMessage(env, addr[0], addr[1], random.randint(0, APP_MAX_DELAY),
                                unicast_num=self._unicast_num)

    def _sendProbe(self, types=None, scopes=None, address=None, port=None):
        env = constructProbe(types, scopes)
        if self._dpActive:
            self.sendUnicastMessage(env, self._dpAddr[0], self._dpAddr[1],
                                    unicast_num=self._unicast_num)
        elif address and port:
            self.sendUnicastMessage(env, address, port,
                                    unicast_num=self._unicast_num)
        else:
            self.sendMulticastMessage(env, multicast_num=self._multicast_num)

    def _sendResolve(self, epr):
        env = constructResolve(epr)
        if self._dpActive:
            self.sendUnicastMessage(env, self._dpAddr[0], self._dpAddr[1],
                                    unicast_num=self._unicast_num)
        else:
            self.sendMulticastMessage(env, multicast_num=self._multicast_num)

    def _sendHello(self, service):
        env = constructHello(service)
        random.seed((int)(time.time() * 1000000))
        self.sendMulticastMessage(env,initialDelay=random.randint(0, APP_MAX_DELAY),
                                  multicast_num=self._multicast_num)

    def _sendBye(self, service):
        env = constructBye(service)
        service.incrementMessageNumber()
        self.sendMulticastMessage(env, multicast_num=self._multicast_num)



import logging
import selectors
import socket
from .fixtures import probe_response
from wsdiscovery.threaded import NetworkingThread, MULTICAST_PORT
from wsdiscovery import WSDiscovery


def test_probing(monkeypatch, probe_response):
    "mock up socket registration, event selection & socket message response"

    sck = None

    def mock_register(selector, rsock, evtmask):
        """get hold of the multicast socket that will send the Probe message,
        when the socket is registered with the selector"""
        global sck
        # The Probe is sent multicast, we use that to identify the right socket.
        # Identification could be done better, but this is enough for now.
        if (rsock.proto == socket.IPPROTO_UDP and
            rsock.getsockname()[1] == MULTICAST_PORT):
            sck = rsock

    def mock_select(*args):
        "set a mock Probe response event in motion for the same socket"
        global sck
        if sck and sck.getsockname()[1] == MULTICAST_PORT:
            key = selectors.SelectorKey(sck, sck.fileno(), [], "")
            # to mock just one response we just nullify the sock
            sck = None
            return [(key, selectors.EVENT_READ)]
        else:
            return []

    def mock_recvfrom(*args):
        return probe_response

    monkeypatch.setattr(selectors.DefaultSelector, "register", mock_register)
    monkeypatch.setattr(selectors.DefaultSelector, "select", mock_select)
    monkeypatch.setattr(socket.socket, "recvfrom", mock_recvfrom)

    # we cannot use a fixture that'd start discovery for us, since the socket
    # selector registration happens at startup time
    wsd = WSDiscovery()
    wsd.start()
    found = wsd.searchServices()

    assert len(found) == 1
    assert probe_response[1] in found[0].getXAddrs()[0]
    assert len(found[0].getScopes()) == 4



"""SOAP & WS-Discovery XML namespaces used by the package"""


#: Addressing namespace
NS_ADDRESSING = "http://schemas.xmlsoap.org/ws/2004/08/addressing"

#: Discovery namespace
NS_DISCOVERY = "http://schemas.xmlsoap.org/ws/2005/04/discovery"

#: SOAP envelope namespace
NS_SOAPENV = "http://www.w3.org/2003/05/soap-envelope"

NS_ADDRESS_ALL = "urn:schemas-xmlsoap-org:ws:2005:04:discovery"
NS_ADDRESS_UNKNOWN = "http://schemas.xmlsoap.org/ws/2004/08/addressing/role/anonymous"

#: :term:`Bye` message namespace
NS_ACTION_BYE = "http://schemas.xmlsoap.org/ws/2005/04/discovery/Bye"

#: :term:`Hello` message namespace
NS_ACTION_HELLO = "http://schemas.xmlsoap.org/ws/2005/04/discovery/Hello"

#: :term:`Probe` message namespace
NS_ACTION_PROBE = "http://schemas.xmlsoap.org/ws/2005/04/discovery/Probe"

#: :term:`Probe match` message namespace
NS_ACTION_PROBE_MATCH = "http://schemas.xmlsoap.org/ws/2005/04/discovery/ProbeMatches"

#: :term:`Resolve` message namespace
NS_ACTION_RESOLVE = "http://schemas.xmlsoap.org/ws/2005/04/discovery/Resolve"

#: :term:`Resolve match` message namespace
NS_ACTION_RESOLVE_MATCH = "http://schemas.xmlsoap.org/ws/2005/04/discovery/ResolveMatches"





"Serialize & parse WS-Discovery Resolve SOAP messages"

import uuid

from ..namespaces import NS_ADDRESSING, NS_DISCOVERY, NS_ACTION_RESOLVE, NS_ADDRESS_ALL
from ..envelope import SoapEnvelope
from ..util import createSkelSoapMessage, getBodyEl, getHeaderEl, addElementWithText, \
                   addEPR, addTypes, addScopes, getDocAsString, getScopes


def constructResolve(epr):
    "construct an envelope that represents a ``Resolve`` message"

    env = SoapEnvelope()
    env.setAction(NS_ACTION_RESOLVE)
    env.setTo(NS_ADDRESS_ALL)
    env.setMessageId(uuid.uuid4().urn)
    env.setEPR(epr)
    return env


def createResolveMessage(env):
    "serialize a SOAP envelope object into a string"

    doc = createSkelSoapMessage(NS_ACTION_RESOLVE)

    bodyEl = getBodyEl(doc)
    headerEl = getHeaderEl(doc)

    addElementWithText(doc, headerEl, "a:MessageID", NS_ADDRESSING, env.getMessageId())
    addElementWithText(doc, headerEl, "a:To", NS_ADDRESSING, env.getTo())

    if len(env.getReplyTo()) > 0:
        addElementWithText(doc, headerEl, "a:ReplyTo", NS_ADDRESSING, env.getReplyTo())

    resolveEl = doc.createElementNS(NS_DISCOVERY, "d:Resolve")
    addEPR(doc, resolveEl, env.getEPR())
    bodyEl.appendChild(resolveEl)

    return getDocAsString(doc)


def parseResolveMessage(dom):
    "parse a XML message into a SOAP envelope object"

    env = SoapEnvelope()
    env.setAction(NS_ACTION_RESOLVE)

    env.setMessageId(dom.getElementsByTagNameNS(NS_ADDRESSING, "MessageID")[0].firstChild.data.strip())

    replyToNodes = dom.getElementsByTagNameNS(NS_ADDRESSING, "ReplyTo")
    if len(replyToNodes) > 0:
        env.setReplyTo(replyToNodes[0].firstChild.data.strip())

    env.setTo(dom.getElementsByTagNameNS(NS_ADDRESSING, "To")[0].firstChild.data.strip())
    env.setEPR(dom.getElementsByTagNameNS(NS_ADDRESSING, "Address")[0].firstChild.data.strip())

    return env





"Serialize & parse WS-Discovery Probe Match SOAP messages"

import uuid
import random
import time

from ..namespaces import NS_ADDRESSING, NS_DISCOVERY, NS_ACTION_PROBE_MATCH, NS_ADDRESS_UNKNOWN
from ..envelope import SoapEnvelope
from ..util import createSkelSoapMessage, getBodyEl, getHeaderEl, addElementWithText, \
                   addTypes, addScopes, getDocAsString, getScopes, _parseAppSequence, \
                   addEPR, getXAddrs, addXAddrs, getTypes, _generateInstanceId


def constructProbeMatch(services, relatesTo):
    "construct an envelope that represents a ``Probe Match`` message"

    env = SoapEnvelope()
    env.setAction(NS_ACTION_PROBE_MATCH)
    env.setTo(NS_ADDRESS_UNKNOWN)
    env.setMessageId(uuid.uuid4().urn)
    random.seed((int)(time.time() * 1000000))
    env.setInstanceId(_generateInstanceId())
    env.setMessageNumber("1")
    env.setRelatesTo(relatesTo)

    prbs =  env.getProbeResolveMatches()
    for srv in services:
        prb = ProbeResolveMatch(srv.getEPR(), srv.getTypes(), srv.getScopes(), \
                                srv.getXAddrs(), str(srv.getMetadataVersion()))
        prbs.append(prb)
    return env


def createProbeMatchMessage(env):
    "serialize a SOAP envelope object into a string"

    doc = createSkelSoapMessage(NS_ACTION_PROBE_MATCH)

    bodyEl = getBodyEl(doc)
    headerEl = getHeaderEl(doc)

    addElementWithText(doc, headerEl, "a:MessageID", NS_ADDRESSING, env.getMessageId())
    addElementWithText(doc, headerEl, "a:RelatesTo", NS_ADDRESSING, env.getRelatesTo())
    addElementWithText(doc, headerEl, "a:To", NS_ADDRESSING, env.getTo())

    appSeqEl = doc.createElementNS(NS_DISCOVERY, "d:AppSequence")
    appSeqEl.setAttribute("InstanceId", env.getInstanceId())
    appSeqEl.setAttribute("MessageNumber", env.getMessageNumber())
    headerEl.appendChild(appSeqEl)

    probeMatchesEl = doc.createElementNS(NS_DISCOVERY, "d:ProbeMatches")
    probeMatches = env.getProbeResolveMatches()
    for probeMatch in probeMatches:
        probeMatchEl = doc.createElementNS(NS_DISCOVERY, "d:ProbeMatch")
        addEPR(doc, probeMatchEl, probeMatch.getEPR())
        addTypes(doc, probeMatchEl, probeMatch.getTypes())
        addScopes(doc, probeMatchEl, probeMatch.getScopes())
        addXAddrs(doc, probeMatchEl, probeMatch.getXAddrs())
        addElementWithText(doc, probeMatchEl, "d:MetadataVersion", NS_DISCOVERY, probeMatch.getMetadataVersion())
        probeMatchesEl.appendChild(probeMatchEl)


    bodyEl.appendChild(probeMatchesEl)

    return getDocAsString(doc)


def parseProbeMatchMessage(dom):
    "parse a XML message into a SOAP envelope object"

    env = SoapEnvelope()
    env.setAction(NS_ACTION_PROBE_MATCH)

    env.setMessageId(dom.getElementsByTagNameNS(NS_ADDRESSING, "MessageID")[0].firstChild.data.strip())
    env.setRelatesTo(dom.getElementsByTagNameNS(NS_ADDRESSING, "RelatesTo")[0].firstChild.data.strip())
    # Even though To is required in WS-Discovery, some devices omit it
    elem = dom.getElementsByTagNameNS(NS_ADDRESSING, "To").item(0)
    if elem:
        env.setTo(elem.firstChild.data.strip())

    _parseAppSequence(dom, env)

    pmNodes = dom.getElementsByTagNameNS(NS_DISCOVERY, "ProbeMatch")
    for node in pmNodes:
        epr = node.getElementsByTagNameNS(NS_ADDRESSING, "Address")[0].firstChild.data.strip()

        types = []
        typeNodes = node.getElementsByTagNameNS(NS_DISCOVERY, "Types")
        if len(typeNodes) > 0:
            types = getTypes(typeNodes[0])

        scopes = []
        scopeNodes = node.getElementsByTagNameNS(NS_DISCOVERY, "Scopes")
        if len(scopeNodes) > 0:
            scopes = getScopes(scopeNodes[0])

        xAddrs = []
        xAddrNodes = node.getElementsByTagNameNS(NS_DISCOVERY, "XAddrs")
        if len(xAddrNodes) > 0:
            xAddrs = getXAddrs(xAddrNodes[0])

        mdv = node.getElementsByTagNameNS(NS_DISCOVERY, "MetadataVersion")[0].firstChild.data.strip()
        env.getProbeResolveMatches().append(ProbeResolveMatch(epr, types, scopes, xAddrs, mdv))

    return env



class ProbeResolveMatch:

    def __init__(self, epr, types, scopes, xAddrs, metadataVersion):
        self._epr = epr
        self._types = types
        self._scopes = scopes
        self._xAddrs = xAddrs
        self._metadataVersion = metadataVersion

    def getEPR(self):
        return self._epr

    def getTypes(self):
        return self._types

    def getScopes(self):
        return self._scopes

    def getXAddrs(self):
        return self._xAddrs

    def getMetadataVersion(self):
        return self._metadataVersion

    def __repr__(self):
        return "EPR: %s\nTypes: %s\nScopes: %s\nXAddrs: %s\nMetadata Version: %s" % \
            (self.getEPR(), self.getTypes(), self.getScopes(),
             self.getXAddrs(), self.getMetadataVersion())






"""Various utilities used by different parts of the package."""

import io
import string
import random
import logging
import ipaddress
import socket
from xml.dom import minidom
import ifaddr

from .scope import Scope
from .uri import URI
from .namespaces import NS_ADDRESSING, NS_DISCOVERY, NS_SOAPENV
from .qname import QName

logger = logging.getLogger("util")

_APIPA_RANGE = ipaddress.ip_network("169.254.0.0/16")


def createSkelSoapMessage(soapAction):
    doc = minidom.Document()

    envEl = doc.createElementNS(NS_SOAPENV, "s:Envelope")

    envEl.setAttribute("xmlns:a", NS_ADDRESSING)  # minidom does not insert this automatically
    envEl.setAttribute("xmlns:d", NS_DISCOVERY)
    envEl.setAttribute("xmlns:s", NS_SOAPENV)

    doc.appendChild(envEl)

    headerEl = doc.createElementNS(NS_SOAPENV, "s:Header")
    envEl.appendChild(headerEl)

    addElementWithText(doc, headerEl, "a:Action", NS_ADDRESSING, soapAction)

    bodyEl = doc.createElementNS(NS_SOAPENV, "s:Body")
    envEl.appendChild(bodyEl)

    return doc


def addElementWithText(doc, parent, name, ns, value):
    el = doc.createElementNS(ns, name)
    text = doc.createTextNode(value)
    el.appendChild(text)
    parent.appendChild(el)


def addEPR(doc, node, epr):
    eprEl = doc.createElementNS(NS_ADDRESSING, "a:EndpointReference")
    addElementWithText(doc, eprEl, "a:Address", NS_ADDRESSING, epr)
    node.appendChild(eprEl)


def addScopes(doc, node, scopes):
    if scopes is not None and len(scopes) > 0:
        addElementWithText(doc, node, "d:Scopes", NS_DISCOVERY, " ".join([x.getQuotedValue() for x in scopes]))
        if scopes[0].getMatchBy() is not None and len(scopes[0].getMatchBy()) > 0:
            node.getElementsByTagNameNS(NS_DISCOVERY, "Scopes")[0].setAttribute("MatchBy", scopes[0].getMatchBy())


def addTypes(doc, node, types):
    if types is not None and len(types) > 0:
        envEl = getEnvEl(doc)
        typeList = []
        prefixMap = {}
        for type in types:
            ns = type.getNamespace()
            localname = type.getLocalname()
            if type.getNamespacePrefix() is None:
                if prefixMap.get(ns) == None:
                    prefix = getRandomStr()
                    prefixMap[ns] = prefix
                else:
                    prefix = prefixMap.get(ns)
            else:
                prefix = type.getNamespacePrefix()
            addNSAttrToEl(envEl, ns, prefix)
            typeList.append(prefix + ":" + localname)
        addElementWithText(doc, node, "d:Types", NS_DISCOVERY, " ".join(typeList))


def addXAddrs(doc, node, xAddrs):
    if xAddrs is not len(xAddrs) > 0:
        addElementWithText(doc, node, "d:XAddrs", NS_DISCOVERY, " ".join([x for x in xAddrs]))


def getDocAsString(doc):
    outStr = None
    stream = io.StringIO(outStr)
    stream.write(doc.toprettyxml())
    return stream.getvalue()


def getBodyEl(doc):
    return doc.getElementsByTagNameNS(NS_SOAPENV, "Body")[0]


def getHeaderEl(doc):
    return doc.getElementsByTagNameNS(NS_SOAPENV, "Header")[0]


def getEnvEl(doc):
    return doc.getElementsByTagNameNS(NS_SOAPENV, "Envelope")[0]


def addNSAttrToEl(el, ns, prefix):
    el.setAttribute("xmlns:" + prefix, ns)


def _parseAppSequence(dom, env):
    nodes = dom.getElementsByTagNameNS(NS_DISCOVERY, "AppSequence")
    if nodes:
        appSeqNode = nodes[0]
        env.setInstanceId(appSeqNode.getAttribute("InstanceId"))
        env.setSequenceId(appSeqNode.getAttribute("SequenceId"))
        env.setMessageNumber(appSeqNode.getAttribute("MessageNumber"))


def _parseSpaceSeparatedList(node):
    if node.childNodes:
        return [item.replace('%20', ' ') \
            for item in node.childNodes[0].data.split()]
    else:
        return []


def extractSoapUdpAddressFromURI(uri):
    val = uri.getPathExQueryFragment().split(":")
    part1 = val[0][2:]
    part2 = None
    if val[1].count('/') > 0:
        part2 = int(val[1][:val[1].index('/')])
    else:
        part2 = int(val[1])
    addr = [part1, part2]
    return addr


def getXAddrs(xAddrsNode):
    return _parseSpaceSeparatedList(xAddrsNode)


def getTypes(typeNode):
    return [getQNameFromValue(item, typeNode) \
                for item in _parseSpaceSeparatedList(typeNode)]


def getScopes(scopeNode):
    matchBy = scopeNode.getAttribute("MatchBy")
    return [Scope(item, matchBy) \
                for item in _parseSpaceSeparatedList(scopeNode)]


def matchScope(src, target, matchBy):

    MATCH_BY_LDAP = "http://schemas.xmlsoap.org/ws/2005/04/discovery/ldap"
    MATCH_BY_URI = "http://schemas.xmlsoap.org/ws/2005/04/discovery/rfc2396"
    MATCH_BY_UUID = "http://schemas.xmlsoap.org/ws/2005/04/discovery/uuid"
    MATCH_BY_STRCMP = "http://schemas.xmlsoap.org/ws/2005/04/discovery/strcmp0"

    if matchBy == "" or matchBy == None or matchBy == MATCH_BY_LDAP or matchBy == MATCH_BY_URI or matchBy == MATCH_BY_UUID:
        src = URI(src)
        target = URI(target)
        if src.getScheme().lower() != target.getScheme().lower():
            return False
        if src.getAuthority().lower() != target.getAuthority().lower():
            return False
        srcPath = src.getPathExQueryFragment()
        targetPath = target.getPathExQueryFragment()
        if srcPath == targetPath:
            return True
        elif targetPath.startswith(srcPath):
            n = len(srcPath)
            if targetPath[n - 1] == srcPath[n - 1] == '/':
                return True
            if targetPath[n] == '/':
                return True
            return False
        else:
            return False
    elif matchBy == MATCH_BY_STRCMP:
        return src == target
    else:
        return False


def isTypeInList(ttype, types):
    for entry in types:
        if ttype.getFullname() == entry.getFullname():
            return True
    return False


def isScopeInList(scope, scopes):
    for entry in scopes:
        if matchScope(scope.getValue(), entry.getValue(), scope.getMatchBy()):
            return True
    return False


def matchesFilter(service, types, scopes):
    if types is not None:
        for ttype in types:
            if not isTypeInList(ttype, service.getTypes()):
                return False
    if scopes is not None:
        for scope in scopes:
            if not isScopeInList(scope, service.getScopes()):
                return False
    return True


def filterServices(services, types, scopes):
    return [service for service in services if matchesFilter(service, types, scopes)]


def getNamespaceValue(node, prefix):
    while node != None:
        if node.nodeType == minidom.Node.ELEMENT_NODE:
            attr = node.getAttributeNode("xmlns:" + prefix)
            if attr != None:
                return attr.nodeValue
        node = node.parentNode
    return ""


def getDefaultNamespace(node):
    while node != None:
        if node.nodeType == minidom.Node.ELEMENT_NODE:
            attr = node.getAttributeNode("xmlns")
            if attr != None:
                return attr.nodeValue
        node = node.parentNode
    return ""


def getQNameFromValue(value, node):
    vals = value.split(":")
    ns = ""
    prefix = None
    if len(vals) == 1:
        localName = vals[0]
        ns = getDefaultNamespace(node)
    else:
        localName = vals[1]
        prefix = vals[0]
        ns = getNamespaceValue(node, prefix)
    return QName(ns, localName, prefix)


def _getNetworkAddrs(protocol_version):
    addrs = []
    ifaces = ifaddr.get_adapters()

    # ifaddr library only returns IPv4 and IPv6 addresses
    if protocol_version == socket.AF_INET:
        for iface in ifaces:
            for ip in iface.ips:
                if isinstance(ip.ip, str):
                    ip_address = ipaddress.ip_address(ip.ip)
                    if not ip_address.is_loopback and ip_address not in _APIPA_RANGE:
                        addrs.append(ip_address)
    elif protocol_version == socket.AF_INET6:
        for iface in ifaces:
            for ip in iface.ips:
                if isinstance(ip.ip, tuple):
                    ip_address = ipaddress.ip_address(f"{ip.ip[0]}%{ip.ip[2]}")
                    if not ip_address.is_loopback and ip_address not in _APIPA_RANGE:
                        addrs.append(ip_address)
    else:
        logger.warning(f"requested protocol version ({protocol_version}) is not"
                       f" IPv4 ({socket.AF_INET}) or IPv6 ({socket.AF_INET6})")

    return addrs


def _generateInstanceId():
    return str(random.randint(1, 0xFFFFFFFF))


def getRandomStr():
    return "".join([random.choice(string.ascii_letters) for x in range(10)])


def showEnv(env):
    print("-----------------------------")
    print("Action: %s" % env.getAction())
    print("MessageId: %s" % env.getMessageId())
    print("InstanceId: %s" % env.getInstanceId())
    print("MessageNumber: %s" % env.getMessageNumber())
    print("Reply To: %s" % env.getReplyTo())
    print("To: %s" % env.getTo())
    print("RelatesTo: %s" % env.getRelatesTo())
    print("Relationship Type: %s" % env.getRelationshipType())
    print("Types: %s" % env.getTypes())
    print("Scopes: %s" % env.getScopes())
    print("EPR: %s" % env.getEPR())
    print("Metadata Version: %s" % env.getMetadataVersion())
    print("Probe Matches: %s" % env.getProbeResolveMatches())
    print("-----------------------------")



def dom2Str(data):
    dom = minidom.parseString(data)
    return "\n" + dom.toprettyxml(indent="  ") + "\n"






"Serialize & parse WS-Discovery Probe SOAP messages"

import uuid
from xml.dom import minidom

from ..namespaces import NS_ADDRESSING, NS_DISCOVERY, NS_ACTION_PROBE, NS_ADDRESS_ALL
from ..envelope import SoapEnvelope
from ..util import createSkelSoapMessage, getBodyEl, getHeaderEl, addElementWithText, \
                   addTypes, getTypes, addScopes, getDocAsString, getScopes


def constructProbe(types, scopes):
    "construct an envelope that represents a ``Probe`` message"

    env = SoapEnvelope()
    env.setAction(NS_ACTION_PROBE)
    env.setTo(NS_ADDRESS_ALL)
    env.setMessageId(uuid.uuid4().urn)
    env.setTypes(types)
    env.setScopes(scopes)
    return env


def createProbeMessage(env):
    "serialize a SOAP envelope object into a string"

    doc = createSkelSoapMessage(NS_ACTION_PROBE)

    bodyEl = getBodyEl(doc)
    headerEl = getHeaderEl(doc)

    addElementWithText(doc, headerEl, "a:MessageID", NS_ADDRESSING, env.getMessageId())
    addElementWithText(doc, headerEl, "a:To", NS_ADDRESSING, env.getTo())

    if len(env.getReplyTo()) > 0:
        addElementWithText(doc, headerEl, "a:ReplyTo", NS_ADDRESSING, env.getReplyTo())

    probeEl = doc.createElementNS(NS_DISCOVERY, "d:Probe")
    bodyEl.appendChild(probeEl)

    addTypes(doc, probeEl, env.getTypes())
    addScopes(doc, probeEl, env.getScopes())

    return getDocAsString(doc)


def parseProbeMessage(dom):
    "parse a XML message into a SOAP envelope object"

    env = SoapEnvelope()
    env.setAction(NS_ACTION_PROBE)
    env.setMessageId(dom.getElementsByTagNameNS(NS_ADDRESSING, "MessageID")[0].firstChild.data.strip())

    replyToNodes = dom.getElementsByTagNameNS(NS_ADDRESSING, "ReplyTo")
    if len(replyToNodes) > 0 and \
       isinstance(replyToNodes[0].firstChild, minidom.Text):
        env.setReplyTo(replyToNodes[0].firstChild.data.strip())

    env.setTo(dom.getElementsByTagNameNS(NS_ADDRESSING, "To")[0].firstChild.data.strip())

    typeNodes = dom.getElementsByTagNameNS(NS_DISCOVERY, "Types")
    if len(typeNodes) > 0:
        env.getTypes().extend(getTypes(typeNodes[0]))

    scopeNodes = dom.getElementsByTagNameNS(NS_DISCOVERY, "Scopes")
    if len(scopeNodes) > 0:
        env.getScopes().extend(getScopes(scopeNodes[0]))

    return env







"""Functions to serialize and deserialize messages between SOAP envelope & string representations"""

import io, sys
from .namespaces import NS_ADDRESSING, NS_SOAPENV
from .actions import *
from xml.dom import minidom


def createSOAPMessage(env):
    "serialize SOAP envelopes into XML strings"

    if env.getAction() == NS_ACTION_PROBE:
        return createProbeMessage(env)
    if env.getAction() == NS_ACTION_PROBE_MATCH:
        return createProbeMatchMessage(env)
    if env.getAction() == NS_ACTION_RESOLVE:
        return createResolveMessage(env)
    if env.getAction() == NS_ACTION_RESOLVE_MATCH:
        return createResolveMatchMessage(env)
    if env.getAction() == NS_ACTION_HELLO:
        return createHelloMessage(env)
    if env.getAction() == NS_ACTION_BYE:
        return createByeMessage(env)


def parseSOAPMessage(data, ipAddr):
    "deserialize XML message strings into SOAP envelope objects"
    try:
        dom = minidom.parseString(data)
    except Exception:
        #print('Failed to parse message from %s\n"%s": %s' % (ipAddr, data, ex), file=sys.stderr)
        return None

    if dom.getElementsByTagNameNS(NS_SOAPENV, "Fault"):
        #print('Fault received from %s %s:' % (ipAddr, data), file=sys.stderr)
        return None

    soapAction = dom.getElementsByTagNameNS(NS_ADDRESSING, "Action")[0].firstChild.data.strip()
    if soapAction == NS_ACTION_PROBE:
        return parseProbeMessage(dom)
    elif soapAction == NS_ACTION_PROBE_MATCH:
        return parseProbeMatchMessage(dom)
    elif soapAction == NS_ACTION_RESOLVE:
        return parseResolveMessage(dom)
    elif soapAction == NS_ACTION_RESOLVE_MATCH:
        return parseResolveMatchMessage(dom)
    elif soapAction == NS_ACTION_BYE:
        return parseByeMessage(dom)
    elif soapAction == NS_ACTION_HELLO:
        return parseHelloMessage(dom)



#!/usr/bin/env python

import os.path

try:
   from setuptools.commands import setup
except:
   from distutils.core import setup

here = os.path.abspath(os.path.dirname(__file__))

try:
    README = open(os.path.join(here, "README.md")).read()
    CHANGES = open(os.path.join(here, "CHANGES.md")).read()
except Exception:
    README = CHANGES = ""


setup(name='WSDiscovery',
      version='2.1.0',
      description='WS-Discovery implementation for python',
      long_description=README + "\n\n" + CHANGES,
      long_description_content_type="text/markdown",
      author='Andrei Kopats',
      author_email='andrei.kopats@gmail.com',
      url='https://github.com/andreikop/python-ws-discovery.git',
      classifiers=[
            'Development Status :: 4 - Beta',
            'Environment :: Console',
            'Intended Audience :: Developers',
            'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)',
            'Operating System :: MacOS :: MacOS X',
            'Operating System :: POSIX :: Linux',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.9',
            'Programming Language :: Python :: 3.10',
            'Programming Language :: Python :: 3.11',
            'Programming Language :: Python :: 3.12',
            'Programming Language :: Python :: 3.13',
            'Topic :: Software Development :: Libraries',
            'Topic :: Software Development :: Libraries :: Python Modules',
            'Topic :: Communications'
      ],
      packages=['wsdiscovery', 'wsdiscovery.actions'],
      setup_requires=['ifaddr', 'click'],
      install_requires=['ifaddr', 'click'],
      tests_require = ['pytest', 'mock'],
      entry_points = {
         'console_scripts': [
            'wsdiscover=wsdiscovery.cmdline:discover',
            'wspublish=wsdiscovery.cmdline:publish'
        ],
      }
     )



""":abbr:`UDP (User Datagram Protocol)` message implementation that helps with
management of unicast/multicast semantics and repeat & delay handling.

WS-Discovery spec dictates that the example algorithm provided in SOAP-over-UDP spec
is to be used; see

http://docs.oasis-open.org/ws-dd/soapoverudp/1.1/os/wsdd-soapoverudp-1.1-spec-os.html#_Toc229451838
"""

import random
import time

# delays are in milliseconds

UNICAST_UDP_REPEAT=2
UNICAST_UDP_MIN_DELAY=50
UNICAST_UDP_MAX_DELAY=250
UNICAST_UDP_UPPER_DELAY=500

MULTICAST_UDP_REPEAT=4
MULTICAST_UDP_MIN_DELAY=50
MULTICAST_UDP_MAX_DELAY=250
MULTICAST_UDP_UPPER_DELAY=500


class UDPMessage:
    "UDP message management implementation"

    UNICAST = 'unicast'
    MULTICAST = 'multicast'

    def __init__(self, env, addr, port, msgType, initialDelay=0,
                 unicast_num=UNICAST_UDP_REPEAT,
                 multicast_num=MULTICAST_UDP_REPEAT):
        """msgType shall be UDPMessage.UNICAST or UDPMessage.MULTICAST"""
        self._env = env
        self._addr = addr
        self._port = port
        self._msgType = msgType

        if msgType == self.UNICAST:
            udpRepeat, udpMinDelay, udpMaxDelay, udpUpperDelay = \
                    unicast_num, \
                    UNICAST_UDP_MIN_DELAY, \
                    UNICAST_UDP_MAX_DELAY, \
                    UNICAST_UDP_UPPER_DELAY
        else:
            udpRepeat, udpMinDelay, udpMaxDelay, udpUpperDelay = \
                    multicast_num, \
                    MULTICAST_UDP_MIN_DELAY, \
                    MULTICAST_UDP_MAX_DELAY, \
                    MULTICAST_UDP_UPPER_DELAY

        self._udpRepeat = udpRepeat
        self._udpUpperDelay = udpUpperDelay
        self._t = (udpMinDelay + ((udpMaxDelay - udpMinDelay) * random.random())) / 2
        self._nextTime = int(time.time() * 1000) + initialDelay

    def getEnv(self):
        return self._env

    def getAddr(self):
        return self._addr

    def getPort(self):
        return self._port

    def msgType(self):
        return self._msgType

    def isFinished(self):
        return self._udpRepeat <= 0

    def canSend(self):
        ct = int(time.time() * 1000)
        return self._nextTime < ct

    def refresh(self):
        self._t = self._t * 2
        if self._t > self._udpUpperDelay:
            self._t = self._udpUpperDelay
        self._nextTime = int(time.time() * 1000) + self._t
        self._udpRepeat = self._udpRepeat - 1



"""Publisher application."""

import logging
import random
import time
import uuid

from .actions import *
from .uri import URI
from .util import filterServices, _generateInstanceId
from .service import Service
from .threaded import ThreadedNetworking
from .daemon import Daemon


class Publishing:
    "networking-agnostic generic service publishing mixin"

    def __init__(self, **kwargs):
        self._localServices = {}
        super().__init__(**kwargs)

    def _handle_probe(self, env, addr):
        "handle NS_ACTION_PROBE"
        services = filterServices(list(self._localServices.values()), env.getTypes(), env.getScopes())
        self._sendProbeMatch(services, env.getMessageId(), addr)

    def _handle_resolve(self, env, addr):
        "handle NS_ACTION_RESOLVE"
        if env.getEPR() in self._localServices:
            service = self._localServices[env.getEPR()]
            self._sendResolveMatch(service, env.getMessageId(), addr)


    def  _networkAddressAdded(self, addr):
        self.addSourceAddr(addr)
        for service in list(self._localServices.values()):
            self._sendHello(service)

    def _networkAddressRemoved(self, addr):
        self.removeSourceAddr(addr)


    def publishService(self, types, scopes, xAddrs):
        """Publish a service with the given TYPES, SCOPES and XAddrs (service addresses)

        if xAddrs contains item, which includes {ip} pattern, one item per IP address will be sent
        """

        if not self._serverStarted:
            raise Exception("Server not started")

        instanceId = _generateInstanceId()

        service = Service(types, scopes, xAddrs, self.uuid, instanceId)
        self._localServices[self.uuid] = service
        self._sendHello(service)


    def clearLocalServices(self):
        'send Bye messages for the services and remove them'

        for service in list(self._localServices.values()):
            self._sendBye(service)

        self._localServices.clear()

    def stop(self):
        self.clearLocalServices()


class ThreadedWSPublishing(ThreadedNetworking, Publishing, Daemon):
    "threaded service publishing"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)



"""Module with URI implementation that supports service scope matching"""

from urllib.parse import unquote


class URI:
    "URI implementation with additional functionality useful for service scope matching"

    def __init__(self, uri):
        uri = unquote(uri)
        i1 = uri.find(":")
        i2 = uri.find("@")
        self._scheme = uri[:i1]
        if i2 != -1:
            self._authority = uri[i1 + 1: i2]
            self._path = uri[i2 + 1:]
        else:
            self._authority = ""
            self._path = uri[i1 + 1:]

    def getScheme(self):
        return self._scheme

    def getAuthority(self):
        return self._authority

    def getPath(self):
        return self._path

    def getPathExQueryFragment(self):
        i = self._path.find("?")
        path = self.getPath()
        if i != -1:
            return path[:self._path.find("?")]
        else:
            return path




"""SOAP envelope implementation."""


class SoapEnvelope:
    "envelope implementation"

    def __init__(self):
        self._action = ""
        self._messageId = ""
        self._relatesTo = ""
        self._relationshipType = None
        self._to = ""
        self._replyTo = ""
        self._instanceId = ""
        self._sequenceId = ""
        self._messageNumber = ""
        self._epr = ""
        self._types = []
        self._scopes = []
        self._xAddrs = []
        self._metadataVersion = "1"
        self._probeResolveMatches = []

    def getAction(self):
        return self._action

    def setAction(self, action):
        self._action = action

    def getMessageId(self):
        return self._messageId

    def setMessageId(self, messageId):
        self._messageId = messageId

    def getRelatesTo(self):
        return self._relatesTo

    def setRelatesTo(self, relatesTo):
        self._relatesTo = relatesTo

    def getRelationshipType(self):
        return self._relationshipType

    def setRelationshipType(self, relationshipType):
        self._relationshipType = relationshipType

    def getTo(self):
        return self._to

    def setTo(self, to):
        self._to = to

    def getReplyTo(self):
        return self._replyTo

    def setReplyTo(self, replyTo):
        self._replyTo = replyTo

    def getInstanceId(self):
        return self._instanceId

    def setInstanceId(self, instanceId):
        self._instanceId = instanceId

    def getSequenceId(self):
        return self._sequenceId

    def setSequenceId(self, sequenceId):
        self._sequenceId = sequenceId

    def getEPR(self):
        return self._epr

    def setEPR(self, epr):
        self._epr = epr

    def getMessageNumber(self):
        return self._messageNumber

    def setMessageNumber(self, messageNumber):
        self._messageNumber = messageNumber

    def getTypes(self):
        return self._types

    def setTypes(self, types):
        self._types = types

    def getScopes(self):
        return self._scopes

    def setScopes(self, scopes):
        self._scopes = scopes

    def getXAddrs(self):
        return self._xAddrs

    def setXAddrs(self, xAddrs):
        self._xAddrs = xAddrs

    def getMetadataVersion(self):
        return self._metadataVersion

    def setMetadataVersion(self, metadataVersion):
        self._metadataVersion = metadataVersion

    def getProbeResolveMatches(self):
        return self._probeResolveMatches

    def setProbeResolveMatches(self, probeResolveMatches):
        self._probeResolveMatches = probeResolveMatches



"""
The actions subpackage provides WS-Discovery action message construction and parsing.
"""

from .bye import NS_ACTION_BYE, constructBye, createByeMessage, parseByeMessage
from .hello import NS_ACTION_HELLO, constructHello, createHelloMessage, parseHelloMessage
from .probe import NS_ACTION_PROBE, constructProbe, createProbeMessage, parseProbeMessage
from .probematch import NS_ACTION_PROBE_MATCH, constructProbeMatch, createProbeMatchMessage, parseProbeMatchMessage
from .probematch import ProbeResolveMatch
from .resolve import NS_ACTION_RESOLVE, constructResolve, createResolveMessage, parseResolveMessage
from .resolvematch import NS_ACTION_RESOLVE_MATCH, constructResolveMatch, createResolveMatchMessage, parseResolveMatchMessage




"""Threaded networking facilities for implementing threaded WS-Discovery daemons."""
import ipaddress
import logging
import platform
import selectors
import socket
import struct
import threading
import time
from typing import cast

from .actions import *
from .message import createSOAPMessage, parseSOAPMessage
from .udp import UDPMessage
from .util import _getNetworkAddrs, dom2Str
from .udp import UNICAST_UDP_REPEAT, MULTICAST_UDP_REPEAT

BUFFER_SIZE = 0xffff
NETWORK_ADDRESSES_CHECK_TIMEOUT = 5
MULTICAST_PORT = 3702
MULTICAST_IPV4_ADDRESS = "239.255.255.250"
MULTICAST_IPV6_ADDRESS = "FF02::C"

logger = logging.getLogger("threading")


class _StoppableDaemonThread(threading.Thread):
    """Stoppable daemon thread.

    run() method shall exit, when self._quitEvent.wait() returned True
    """

    def __init__(self):
        self._quitEvent = threading.Event()
        super(_StoppableDaemonThread, self).__init__()
        self.daemon = True

    def schedule_stop(self):
        """Schedule stopping the thread.
        Use join() to wait, until thread really has been stopped
        """
        self._quitEvent.set()


class AddressMonitorThread(_StoppableDaemonThread):
    "trigger address change callbacks when local service addresses change"

    def __init__(self, wsd, protocol_version):
        self._addrs = set()
        self._wsd = wsd
        self._protocolVersion = protocol_version
        super(AddressMonitorThread, self).__init__()
        self._updateAddrs()

    def _updateAddrs(self):
        addrs = set(_getNetworkAddrs(self._protocolVersion))

        disappeared = self._addrs.difference(addrs)
        new = addrs.difference(self._addrs)

        for addr in disappeared:
            self._wsd._networkAddressRemoved(addr)

        for addr in new:
            self._wsd._networkAddressAdded(addr)

        self._addrs = addrs

    def run(self):
        while not self._quitEvent.wait(NETWORK_ADDRESSES_CHECK_TIMEOUT):
            self._updateAddrs()


class NetworkingThread(_StoppableDaemonThread):
    def __init__(self, observer):
        super(NetworkingThread, self).__init__()

        self.daemon = True
        self._queue = []  # FIXME synchronisation

        self._knownMessageIds = set()
        self._iidMap = {}
        self._observer = observer
        self._capture = observer._capture
        self._relates_to = observer._relates_to

        self._seqnum = 1  # capture sequence number
        self._selector = selectors.DefaultSelector()

    def _makeMreq(self, addr) -> bytes:
        pass

    def _get_inet(self) -> int:
        pass

    def _get_multicast(self) -> int:
        pass

    def _get_ip_proto(self) -> int:
        pass

    def _get_ip_join(self) -> int:
        pass

    def _get_ip_leave(self) -> int:
        pass

    def _get_multicast_ttl(self) -> int:
        pass

    def _createMulticastOutSocket(self, addr, ttl):
        ip_proto = self._get_ip_proto()
        sock = socket.socket(self._get_inet(), socket.SOCK_DGRAM)
        sock.setblocking(0)
        sock.setsockopt(ip_proto, self._get_multicast_ttl(), ttl)

        if not addr:
            iface = socket.INADDR_ANY
        elif self._get_inet() == socket.AF_INET:
            iface = addr.packed
        else:
            iface = int(addr.scope_id)

        sock.setsockopt(ip_proto, self._get_multicast(), iface)

        return sock

    def _createMulticastInSocket(self):
        sock = socket.socket(self._get_inet(), socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

        if platform.system() in ["Darwin", "FreeBSD"]:
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)

        sock.bind(('', MULTICAST_PORT))
        sock.setblocking(0)

        return sock

    def addSourceAddr(self, addr):
        """None means 'system default'"""
        try:
            self._multiInSocket.setsockopt(self._get_ip_proto(), self._get_ip_join(), self._makeMreq(addr))
        except socket.error as e:
            logger.warning(f"Interface has more than 1 address: {e}")

        sock = self._createMulticastOutSocket(addr, self._observer.ttl)
        self._multiOutUniInSockets[addr] = sock
        self._selector.register(sock, selectors.EVENT_READ)

    def removeSourceAddr(self, addr):
        try:
            self._multiInSocket.setsockopt(self._get_ip_proto(), self._get_ip_leave(), self._makeMreq(addr))
        except socket.error as e:
            logger.warning(f"Interface has more than 1 address: {e}")

        sock = self._multiOutUniInSockets[addr]
        self._selector.unregister(sock)
        sock.close()
        del self._multiOutUniInSockets[addr]

    def addUnicastMessage(self, env, addr, port, initialDelay=0,
                          unicast_num=UNICAST_UDP_REPEAT):
        msg = UDPMessage(env, addr, port, UDPMessage.UNICAST, initialDelay,
                         unicast_num=unicast_num)

        self._queue.append(msg)
        self._knownMessageIds.add(env.getMessageId())

    def addMulticastMessage(self, env, addr, port, initialDelay=0,
                            multicast_num=MULTICAST_UDP_REPEAT):
        msg = UDPMessage(env, addr, port, UDPMessage.MULTICAST, initialDelay,
                         multicast_num=multicast_num)

        self._queue.append(msg)
        self._knownMessageIds.add(env.getMessageId())

    def run(self):
        if self._capture:
            self.t0 = time.time()
        while not self._quitEvent.is_set() or self._queue:
            self._sendPendingMessages()
            self._recvMessages()

    def _recvMessages(self):
        for key, events in self._selector.select(0):
            if self._quitEvent.is_set():
                break

            sock = cast(socket.socket, key.fileobj)

            try:
                data, addr = sock.recvfrom(BUFFER_SIZE)
            except socket.error:
                time.sleep(0.01)
                continue

            env = parseSOAPMessage(data, addr[0])

            if env is None:  # fault or failed to parse
                if self._capture:
                    self._capture.write(
                        "%i WARNING: BAD RECV %s:%s TS=%s\n" % (self._seqnum, addr[0], addr[1], time.time() - self.t0))
                    self._capture.write(dom2Str(data))
                    self._seqnum += 1
                continue

            _own_addrs = self._observer._addrsMonitorThread_v4._addrs
            if addr[0] not in _own_addrs:
                if env.getAction() == NS_ACTION_PROBE_MATCH:
                    prms = "\n ".join((str(prm) for prm in env.getProbeResolveMatches()))
                    msg = "probe response from %s:\n --- begin ---\n%s\n--- end ---\n"
                    logger.debug(msg, addr[0], prms)

                if self._capture:
                    self._capture.write(
                        "%i RECV %s:%s TS=%s\n" % (self._seqnum, addr[0], addr[1], time.time() - self.t0))
                    self._capture.write(dom2Str(data))
                    self._seqnum += 1

            mid = env.getMessageId()
            if mid in self._knownMessageIds:
                if self._relates_to and env.getRelatesTo() in self._knownMessageIds:
                    pass
                else:
                    continue
            else:
                if self._capture:
                    self._capture.write("NEW KNOWN MSG IDS %s\n" % (mid))
                self._knownMessageIds.add(mid)

            iid = env.getInstanceId()
            if len(iid) > 0 and int(iid) >= 0:
                mnum = env.getMessageNumber()
                key = addr[0] + ":" + str(addr[1]) + ":" + str(iid)
                if mid is not None and len(mid) > 0:
                    key = key + ":" + mid
                if key not in self._iidMap:
                    self._iidMap[key] = iid
                else:
                    tmnum = self._iidMap[key]
                    if mnum > tmnum:
                        self._iidMap[key] = mnum
                    else:
                        continue

            self._observer.envReceived(env, addr)

    def _sendMsg(self, msg):
        data = createSOAPMessage(msg.getEnv()).encode("UTF-8")

        if msg.msgType() == UDPMessage.UNICAST:
            self._uniOutSocket.sendto(data, (msg.getAddr(), msg.getPort()))
            if self._capture:
                self._capture.write(
                    "%i SEND %s:%s TS=%s\n" % (self._seqnum, msg.getAddr(), msg.getPort(), time.time() - self.t0))
                self._capture.write(dom2Str(data))
                self._seqnum += 1
        else:
            for addr, sock in self._multiOutUniInSockets.items():
                try:
                    sock.sendto(data, (msg.getAddr(), msg.getPort()))
                except OSError as e:
                    # sendto will fail for interfaces that do not support multicast or are not up.
                    # An example of the first case is a wireguard vpn interface.
                    # In either case just log as debug and ignore the error.
                    logger.debug("Interface for %s does not support multicast or is not UP.\n\tOSError %s",
                                 socket.inet_ntoa(sock.getsockopt(self._get_ip_proto(), self._get_multicast(), 4)), e)
                if self._capture:
                    self._capture.write("%i SEND %s:%s iface=%s TS=%s\n" % (
                        self._seqnum, msg.getAddr(), msg.getPort(), addr, time.time() - self.t0))
                    self._capture.write(dom2Str(data))
                    self._seqnum += 1

    def _sendPendingMessages(self):
        """Method sleeps, if nothing to do"""
        if len(self._queue) == 0:
            time.sleep(0.1)
            return
        msg = self._queue.pop(0)
        if msg.canSend():
            self._sendMsg(msg)
            msg.refresh()
            if not (msg.isFinished()):
                self._queue.append(msg)
        else:
            self._queue.append(msg)
            time.sleep(0.01)

    def start(self):
        super(NetworkingThread, self).start()

        self._uniOutSocket = socket.socket(self._get_inet(), socket.SOCK_DGRAM)

        self._multiInSocket = self._createMulticastInSocket()
        self._selector.register(self._multiInSocket, selectors.EVENT_WRITE | selectors.EVENT_READ)

        self._multiOutUniInSockets = {}  # FIXME synchronisation

    def join(self, **kwargs):
        assert self._quitEvent.is_set()
        super(NetworkingThread, self).join()

        self._selector.unregister(self._multiInSocket)
        self._uniOutSocket.close()
        self._multiInSocket.close()

        for sock in self._multiOutUniInSockets.values():
            try:
                sock.close()
            except socket.error as e:
                logger.error(e)


class NetworkingThreadIPv4(NetworkingThread):
    def __init__(self, observer):
        super().__init__(observer)

    def _makeMreq(self, addr):
        "IPvv multicast group join/leave request"
        return struct.pack("4s4s", socket.inet_aton(MULTICAST_IPV4_ADDRESS), addr.packed)

    def _get_inet(self):
        return socket.AF_INET

    def _get_multicast(self):
        return socket.IP_MULTICAST_IF

    def _get_ip_proto(self):
        return socket.IPPROTO_IP

    def _get_ip_join(self):
        return socket.IP_ADD_MEMBERSHIP

    def _get_ip_leave(self):
        return socket.IP_DROP_MEMBERSHIP

    def _get_multicast_ttl(self):
        return socket.IP_MULTICAST_TTL


class NetworkingThreadIPv6(NetworkingThread):
    def __init__(self, observer):
        super().__init__(observer)

    def _makeMreq(self, addr):
        "IPv6 multicast group join/leave request"
        return struct.pack("=16si", socket.inet_pton(socket.AF_INET6, MULTICAST_IPV6_ADDRESS), int(addr.scope_id))

    def _get_inet(self):
        return socket.AF_INET6

    def _get_multicast(self):
        return socket.IPV6_MULTICAST_IF

    def _get_ip_proto(self):
        return socket.IPPROTO_IPV6

    def _get_ip_join(self):
        return socket.IPV6_JOIN_GROUP

    def _get_ip_leave(self):
        return socket.IPV6_LEAVE_GROUP

    def _get_multicast_ttl(self):
        return socket.IPV6_MULTICAST_HOPS


class ThreadedNetworking:
    "handle threaded networking start & stop, address add/remove & message sending"

    def __init__(self,
                 unicast_num=UNICAST_UDP_REPEAT,
                 multicast_num=MULTICAST_UDP_REPEAT,
                 relates_to=False, **kwargs):
        self._networkingThread_v4 = None
        self._networkingThread_v6 = None
        self._addrsMonitorThread_v4 = None
        self._addrsMonitorThread_v6 = None
        self._serverStarted = False
        self._unicast_num = unicast_num
        self._multicast_num = multicast_num
        self._relates_to = relates_to
        super().__init__(**kwargs)

    def _startThreads(self):
        if self._networkingThread_v4 is not None:
            return

        self._networkingThread_v4 = NetworkingThreadIPv4(self)
        self._networkingThread_v6 = NetworkingThreadIPv6(self)
        self._networkingThread_v4.start()
        self._networkingThread_v6.start()
        logger.debug("networking threads started")

        self._addrsMonitorThread_v4 = AddressMonitorThread(self, socket.AF_INET)
        self._addrsMonitorThread_v6 = AddressMonitorThread(self, socket.AF_INET6)
        self._addrsMonitorThread_v4.start()
        self._addrsMonitorThread_v6.start()
        logger.debug("address monitoring threads started")

    def _stopThreads(self):
        if self._networkingThread_v4 is None:
            return

        self._networkingThread_v4.schedule_stop()
        self._addrsMonitorThread_v4.schedule_stop()
        self._networkingThread_v6.schedule_stop()
        self._addrsMonitorThread_v6.schedule_stop()

        self._networkingThread_v4.join()
        self._addrsMonitorThread_v4.join()
        self._networkingThread_v6.join()
        self._addrsMonitorThread_v6.join()

        self._networkingThread_v4 = None
        self._networkingThread_v6 = None

    def start(self):
        """start networking - should be called before using other methods"""
        self._startThreads()
        self._serverStarted = True

    def stop(self):
        """cleans up and stops networking"""
        self._stopThreads()
        self._serverStarted = False

    def addSourceAddr(self, addr):
        version = ipaddress.ip_address(addr).version
        if version == 4:
            self._networkingThread_v4.addSourceAddr(addr)
        elif version == 6:
            self._networkingThread_v6.addSourceAddr(addr)

    def removeSourceAddr(self, addr):
        version = ipaddress.ip_address(addr).version
        if version == 4:
            self._networkingThread_v4.removeSourceAddr(addr)
        elif version == 6:
            self._networkingThread_v6.removeSourceAddr(addr)

    def sendUnicastMessage(self, env, host, port, initialDelay=0,
                           unicast_num=UNICAST_UDP_REPEAT):
        "handle unicast message sending"
        self._networkingThread_v4.addUnicastMessage(env, host, port,
                                                    initialDelay, unicast_num)
        self._networkingThread_v6.addUnicastMessage(env, host, port,
                                                    initialDelay, unicast_num)

    def sendMulticastMessage(self, env, initialDelay=0,
                             multicast_num=MULTICAST_UDP_REPEAT):
        "handle multicast message sending"
        self._networkingThread_v4.addMulticastMessage(env,
                                                      MULTICAST_IPV4_ADDRESS,
                                                      MULTICAST_PORT,
                                                      initialDelay,
                                                      multicast_num)
        self._networkingThread_v6.addMulticastMessage(env,
                                                      MULTICAST_IPV6_ADDRESS,
                                                      MULTICAST_PORT,
                                                      initialDelay,
                                                      multicast_num)



import sys
import logging
from contextlib import contextmanager
from urllib.parse import urlparse
import click

from wsdiscovery.discovery import ThreadedWSDiscovery as WSDiscovery
from wsdiscovery.publishing import ThreadedWSPublishing as WSPublishing
from wsdiscovery.scope import Scope
from wsdiscovery.qname import QName
from wsdiscovery.discovery import DEFAULT_DISCOVERY_TIMEOUT
from wsdiscovery.udp import UNICAST_UDP_REPEAT, MULTICAST_UDP_REPEAT

DEFAULT_LOGLEVEL = "INFO"

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


@contextmanager
def discovery(capture=None, unicast_num=UNICAST_UDP_REPEAT,
              multicast_num=MULTICAST_UDP_REPEAT, relates_to=False):
    wsd = WSDiscovery(capture=capture, unicast_num=unicast_num,
                      multicast_num=multicast_num, relates_to=relates_to)
    wsd.start()
    yield wsd
    wsd.stop()

@contextmanager
def publishing(capture=None):
    wsd = WSPublishing(capture=capture)
    wsd.start()
    yield wsd
    wsd.stop()


def setup_logger(name, loglevel):
    level = getattr(logging, loglevel, None)
    if not level:
        print("Invalid log level '%s'" % loglevel)
        sys.exit()

    logging.basicConfig(level=level)
    return logging.getLogger(name)


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option('--scope', '-s', help='Probe scope URI, e.g. onvif://www.onvif.org/Model/')
@click.option('--type', '-y', 'ptype', nargs=3, type=(str, str, str), multiple=True,
              help='Probe type in this order: NS_URI NS_PREFIX LOCAL_NAME,'
                   ' e.g. http://www.onvif.org/ver10/network/wsdl dp0 NetworkVideoTransmitter'
                   ' -- this option can be specified multiple times')
@click.option('--address', '-a', help='Service address')
@click.option('--port', '-p', type=int, help='Service port')
@click.option('--loglevel', '-l',  default=DEFAULT_LOGLEVEL, show_default=True,
              type=click.Choice(["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]),
              help='Log level')
@click.option('--capture', '-c', nargs=1, type=click.File('w'), help='Capture messages to a file')
@click.option('--timeout', '-t', default=DEFAULT_DISCOVERY_TIMEOUT, show_default=True,
              type=int, help='Discovery timeout in seconds')
@click.option('--unicast-num', '-un', type=int, default=UNICAST_UDP_REPEAT,
              show_default=True, help='Number of Unicast messages to send')
@click.option('--multicast-num', '-mn', type=int, default=MULTICAST_UDP_REPEAT,
              show_default=True, help='Number of Multicast messages to send')
@click.option('--relates-to', '-rt', is_flag=True,
              help='Also use RelatesTo tag to recognize incoming messages.')
def discover(scope, ptype, address, port, loglevel, capture, timeout,
             unicast_num, multicast_num, relates_to):
    "Discover services using WS-Discovery"

    logger = setup_logger("ws-discovery", loglevel)

    probe_types = []
    for type_tuple in ptype:
        probe_types.append(QName(type_tuple[0], type_tuple[2], type_tuple[1]))
    if len(probe_types) == 0:
        probe_types = None

    with discovery(capture, unicast_num, multicast_num, relates_to) as wsd:
        scopes = [Scope(scope)] if scope else []
        svcs = wsd.searchServices(types=probe_types, scopes=scopes,
                                  address=address, port=port, timeout=timeout)
        print("\nDiscovered:\n")
        for service in svcs:
            url = urlparse(service.getXAddrs()[0])
            print(" address: %s" % url.netloc)
            print("  - %s\n" % "\n  - ".join([str(s) for s in service.getScopes()]))


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option('--scope', '-s', help='Full scope URI, eg. onvif://www.onvif.org/Model/')
@click.option('--typename', '-t', help='Qualified type name, eg. https://myservicesns:myservice_type')
@click.option('--address', '-a', help='Service IP address')
@click.option('--port', '-p', type=int, help='Service port')
@click.option('--loglevel', '-l',  default=DEFAULT_LOGLEVEL, show_default=True,
              type=click.Choice(["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]),
              help='Log level')
@click.option('--capture', '-c', nargs=1, type=click.File('w'), help='Capture messages to a file')
@click.option('--unicast-num', '-un', type=int, default=UNICAST_UDP_REPEAT,
              show_default=True, help='Number of Unicast messages to send')
@click.option('--multicast-num', '-mn', type=int, default=MULTICAST_UDP_REPEAT,
              show_default=True, help='Number of Multicast messages to send')
def publish(scope, typename, address, port, loglevel, capture, unicast_num,
            multicast_num):
    "Publish services using WS-Discovery"

    logger = setup_logger("ws-publishing", loglevel)

    with publishing(capture) as wsp:
        scopes = [Scope(scope)] if scope else []

        try:
            proto, ns, name = typename.split(':')
        except:
            types = []
        else:
            ns = ns[2:]
            types = [QName(proto, ns)]

        xAddrs = ["%s:%i" % (address, port)] if address else ['127.0.0.1']
        svc = wsp.publishService(types, scopes, xAddrs)
